@@ -1,4 +1,4 @@
 #!/bin/sh
 
 /scripts/add_default_route.sh &
-exec srsue $*
+exec srsue "$@"
